from distutils.core import setup
setup(
  name = 'emitter.py',
  packages = ['emitter'],
  version = '6.0.0',
  description = 'A minimalist event emitter for Python 3.',
  author = 'titouandk',
  author_email = 'dk.titouan@gmail.com',
  url = 'https://github.com/titouandk/emitter.py',
  download_url = 'https://github.com/titouandk/emitter.py/tarball/6.0.0',
  keywords = ['event emitter', 'event', 'emitter', 'python 3', 'python3'],
  classifiers = ['Programming Language :: Python :: 3'],
)
